from socketIO_client import SocketIO


class AtlasStream(object):

    CHANNEL_RESULT = "atlas_result"
    CHANNEL_PROBE = "atlas_probe"
    CHANNEL_ERROR = "atlas_error"
    CHANNELS = {
        "result": CHANNEL_RESULT,
        "probe": CHANNEL_PROBE,
        "error": CHANNEL_ERROR,
    }

    def __init__(self):
        """Initialize stream"""

        self.iosocket_server = "atlas-stream.ripe.net"
        self.iosocket_resource = "/stream/socket.io"

        self.socketIO = None

    def connect(self):
        """Initiate the channel we want to start streams from."""
        self.socketIO = SocketIO(
            host=self.iosocket_server,
            port=80,
            resource=self.iosocket_resource,
            transports=["websocket"]
        )

    def disconnect(self):
        """Exits the channel k shuts down connection."""
        self.socketIO.disconnect()
        self.socketIO.__exit__([])

    def bind_stream(self, stream_type, callback):
        """Bind given type stream with the given callback"""
        try:
            self.socketIO.on(self.CHANNELS[stream_type], callback)
        except KeyError:
            print("The given stream type: <{0}> is not valid".format(stream_type))

    def start_stream(self, stream_type, **stream_parameters):
        """Starts new stream for given type with given parameters"""
        if stream_type in ("result", "probestatus"):
            self.subscribe(stream_type, **stream_parameters)
        else:
            print("Given stream type: <{0}> is not valid".format(stream_type))

    def subscribe(self, stream_type, **parameters):
        """Subscribe to stream with give parameters."""
        parameters.update({"stream_type": stream_type})
        self.socketIO.emit('atlas_subscribe', parameters)

    def timeout(self, seconds=None):
        """
        Times out all streams after n seconds or wait forever if seconds is
        None
        """
        if seconds is None:
            self.socketIO.wait()
        else:
            self.socketIO.wait(seconds=seconds)



class CousteauGenericError(Exception):
    """Custom Exception class for cousteau general erorrs."""
    pass


class APIResponseError(Exception):
    """Custom Exception class for errors in ATLAS API responses."""
    pass



"""
Module containing the main class that create the post data and makes the HTTP
request according to the ATLAS API.
"""

import calendar
import requests
from dateutil import parser
from datetime import datetime

from .version import __version__


class AtlasRequest(object):
    """
    Base class for doing Atlas requests. Contains functions that can be used by
    most Atlas requests.
    """

    http_methods = {
        "GET": requests.get,
        "POST": requests.post,
        "DELETE": requests.delete
    }

    def __init__(self, **kwargs):
        self.url = ""
        self.url_path = ""
        self.key = kwargs.get("key", "")
        if "url_path" in kwargs:
            self.url_path = kwargs["url_path"]
        if "server" in kwargs:
            self.server = kwargs["server"]
        else:
            self.server = "atlas.ripe.net"

        self.http_agent = "RIPE ATLAS Cousteau v{0}".format(__version__)

        self.http_method_args = {
            "params": {"key": self.key},
            "headers": self.get_headers()
        }
        self.post_data = {}

    def get_headers(self):
        """Return header for the HTTP request."""
        return {
            "User-Agent": self.http_agent,
            "Content-Type": "application/json",
            "Accept": "application/json"
        }

    def http_method(self, method):
        """
        Execute the given HTTP method and returns if it's success or not
        and the response as a string if not success and as python object after
        unjson if it's success.
        """
        return_json_methods = ["GET", "POST"]
        self.build_url()

        try:
            response = self.get_http_method(method)
            is_success = response.ok
            if is_success and method in return_json_methods:
                response_message = response.json()
            else:
                response_message = response.text
        except requests.exceptions.RequestException as exc:
            is_success = False
            response_message = exc.args

        return is_success, response_message

    def get_http_method(self, method):
        """Gets the http method that will be called from the requests library"""
        return self.http_methods[method](self.url, **self.http_method_args)

    def build_url(self):
        """
        Builds the request's url combining server and url_path
        classes attributes.
        """
        self.url = "https://{0}{1}".format(self.server, self.url_path)

    def get(self, **url_params):
        """
        Makes the HTTP GET to the url.
        """
        if url_params:
            self.http_method_args["params"].update(url_params)
        return self.http_method("GET")

    def post(self):
        """
        Makes the HTTP POST to the url sending post_data.
        """
        self._construct_post_data()

        post_args = {"json": self.post_data}
        self.http_method_args.update(post_args)

        return self.http_method("POST")

    def _construct_post_data(self):
        raise NotImplementedError


class AtlasCreateRequest(AtlasRequest):
    """
    Class responsible for creating a request for creating a new Atlas
    measurement. Takes as arguments Atlas API key, a list of Atlas measurement
    objects and a list of Atlas sources. Optionally the start and end time and
    whether the measurement is a oneoff can be specified.
    Usage:
        from ripe.atlas import AtlasCreateRequest
        ar = AtlasCreateRequest(**{
            "start_time": start,
            "stop_time": stop,
            "key": "path_to_key",
            "measurements":[measurement1, ...],
            "sources": [source1, ...],
            "is_oneoff": True/False
        })
        ar.create()
    """

    def __init__(self, **kwargs):
        super(AtlasCreateRequest, self).__init__(**kwargs)

        self.url_path = '/api/v2/measurements/'

        self.measurements = kwargs["measurements"]
        self.sources = kwargs["sources"]
        if kwargs.get("start_time"):
            self.start_time = kwargs["start_time"]
        else:
            self.start_time = ""
        if kwargs.get("stop_time"):
            self.stop_time = kwargs["stop_time"]
        else:
            self.stop_time = ""
        if kwargs.get("is_oneoff"):
            self.is_oneoff = kwargs["is_oneoff"]
        else:
            self.is_oneoff = False

    def _construct_post_data(self):
        """
        Constructs the data structure that is required from the atlas API based
        on measurements, sources and times user has specified.
        """
        definitions = [msm.build_api_struct() for msm in self.measurements]
        probes = [source.build_api_struct() for source in self.sources]
        self.post_data = {
            "definitions": definitions,
            "probes": probes,
            "is_oneoff": self.is_oneoff
        }
        if self.start_time:
            self.post_data.update(
                {"start_time": int(calendar.timegm(self.start_time.timetuple()))}
            )
        if self.stop_time:
            self.post_data.update(
                {"stop_time": int(calendar.timegm(self.stop_time.timetuple()))}
            )

    def create(self):
        """Sends the POST request"""
        return self.post()


class AtlasChangeRequest(AtlasRequest):
    """Atlas request for changing probes for a running measurement.
    post_data = [{
        "action": "add|remove",
        "requested": probe_number,
        # when action=remove only probes is supported
        "type": "area|country|asn|prefix|msm|probes",
        "value": probe_values
    }]
    """

    def __init__(self, **kwargs):
        super(AtlasChangeRequest, self).__init__(**kwargs)

        self.url_path = '/api/v2/measurements/{0}/participation-requests/'
        self.msm_id = kwargs["msm_id"]
        self.sources = kwargs["sources"]
        self.url_path = self.url_path.format(self.msm_id)

    def _construct_post_data(self):
        """
        Constructs the data structure that is required from the atlas API based
        on measurement id, and the sources.
        """
        self.post_data = [source.build_api_struct() for source in self.sources]

    def create(self):
        """Sends the POST request"""
        return self.post()


class AtlasStopRequest(AtlasRequest):
    """Atlas request for stopping a measurement."""

    def __init__(self, **kwargs):
        super(AtlasStopRequest, self).__init__(**kwargs)

        self.url_path = '/api/v2/measurements/'
        self.msm_id = kwargs["msm_id"]
        self.url_path = "{0}{1}".format(self.url_path, self.msm_id)

    def delete(self):
        """
        Makes the HTTP DELETE to the url.
        """
        return self.http_method("DELETE")

    def create(self):
        """Sends the DELETE request"""
        return self.delete()


class AtlasLatestRequest(AtlasRequest):

    def __init__(self, msm_id=None, probe_ids=(), **kwargs):
        super(AtlasLatestRequest, self).__init__(**kwargs)

        self.url_path = "/api/v2/measurements/{0}/latest"

        self.msm_id = msm_id
        self.probe_ids = probe_ids

        self.url_path = self.url_path.format(self.msm_id)
        if self.probe_ids:
            additional_params = {
                "probe_ids": ",".join(map(str, self.probe_ids))
            }

            self.http_method_args["params"].update(additional_params)


class AtlasResultsRequest(AtlasRequest):
    """Atlas request for fetching results of a measurement."""

    def __init__(self, **kwargs):
        super(AtlasResultsRequest, self).__init__(**kwargs)

        self.url_path = '/api/v2/measurements/{0}/results'
        self.msm_id = kwargs["msm_id"]
        start = kwargs.get("start")
        stop = kwargs.get("stop")
        self.probe_ids = kwargs.get("probe_ids")

        self.start = None
        self.stop = None
        self.clean_start_time(start)
        self.clean_stop_time(stop)

        self.url_path = self.url_path.format(self.msm_id)

        self.update_http_method_params()

    def clean_start_time(self, start):
        """
        Transform start time filter to datetime object if there is any.
        """
        if isinstance(start, datetime):
            self.start = start
        elif isinstance(start, int):
            self.start = datetime.utcfromtimestamp(start)
        elif isinstance(start, str):
            self.start = parser.parse(start)

    def clean_stop_time(self, stop):
        """
        Transform stop time filter to datetime object if there is any.
        """
        if isinstance(stop, datetime):
            self.stop = stop
        elif isinstance(stop, int):
            self.stop = datetime.utcfromtimestamp(stop)
        elif isinstance(stop, str):
            self.stop = parser.parse(stop)

    def update_http_method_params(self):
        """
        Update HTTP url parameters based on msm_id and query filters if
        there are any.
        """
        url_params = {}

        if self.start:
            url_params.update(
                {"start": int(calendar.timegm(self.start.timetuple()))}
            )

        if self.stop:
            url_params.update(
                {"stop": int(calendar.timegm(self.stop.timetuple()))}
            )

        if self.probe_ids:
            url_params.update({"probe_ids": ",".join(map(str, self.probe_ids))})

        self.http_method_args["params"].update(url_params)

    def create(self):
        """Sends the GET request."""
        return self.get()


__all__ = [
    "AtlasStopRequest", "AtlasCreateRequest",
    "AtlasChangeRequest", "AtlasRequest",
    "AtlasResultsRequest"
]



"""
Module containing class responsible for creating different Atlas
source objects
"""


class AtlasSource(object):
    """
    Class responsible for creating an Atlas source object that holds
    information about the number of probes, the type of the source and
    the value of the source. This object can be passed as source argument
    later on when we call AtlasRequest.
    Usage:
        from ripe.atlas.cousteau import AtlasSource
        as = AtlasSource(**{"type": "area", "value": "WW", "requested": 5})
    """

    # available types
    types_available = ["area", "country", "prefix", "asn", "probes", "msm"]

    def __init__(self, **kwargs):
        if "requested" in kwargs:
            self.requested = kwargs["requested"]
        else:
            self._requested = None
        if "value" in kwargs:
            self.value = kwargs["value"]
        else:
            self._value = None
        if "type" in kwargs:
            self.type = kwargs["type"]
        else:
            self._type = None
        if "tags" in kwargs:
            self.tags = kwargs["tags"]
        else:
            self._tags = None

    # requested attribute
    def get_requested(self):
        """Getter for requested attribute"""
        return self._requested

    def set_requested(self, value):
        """Setter for requested attribute"""
        self._requested = value

    doc_req = "Defines how many probes will be requested."
    requested = property(get_requested, set_requested, doc=doc_req)

    # value attribute
    def get_value(self):
        """Getter for value attribute"""
        return self._value

    def set_value(self, value):
        """Setter for value attribute"""
        self._value = value

    doc_value = "Defines the value of the type of probe's source."
    value = property(get_value, set_value, doc=doc_value)

    # type attribute
    def get_type(self):
        """Getter for type attribute"""
        return self._type

    def set_type(self, value):
        """Setter for type attribute"""
        if value not in self.types_available:
            log = "Sources field 'type' should be in one of %s" % (
                self.types_available
            )
            raise MalFormattedSource(log)
        self._type = value

    doc_type = "Defines the type of probe's source."
    type = property(get_type, set_type, doc=doc_type)

    # tags attribute
    def get_tags(self):
        """Getter for tags attribute"""
        return self._tags

    def set_tags(self, value):
        """Setter for tags attribute"""
        log = (
            'Sources fields "tags" should be a dict in the format '
            '{"include": [ "tag1", "tag2", "tagN" ],'
            '"exclude": [ "tag1", "tag2", "tagN" ] }'
        )

        if not isinstance(value, dict):
            raise MalFormattedSource(log)

        if not set(value.keys()).issubset(set(["include", "exclude"])):
            raise MalFormattedSource(log)

        for tag_list in value.values():
            if not isinstance(tag_list, list):
                raise MalFormattedSource(log)
            if [tag for tag in tag_list if not isinstance(tag, str)]:
                raise MalFormattedSource(log)

        self._tags = value

    doc_tags = "Defines optional tags to filter probes."
    tags = property(get_tags, set_tags, doc=doc_tags)

    def clean(self):
        """
        Cleans/checks user has entered all required attributes. This might save
        some queries from being sent to server if they are totally wrong.
        """
        if not all([self._requested, self._value, self._type]):
            raise MalFormattedSource(
                "<requested, value, type> fields are required."
            )

    def build_api_struct(self):
        """
        Calls the clean method of the class and returns the info in a structure
        that Atlas API is accepting.
        """
        self.clean()
        r = {
            "type": self._type,
            "requested": self._requested,
            "value": self._value
        }
        if self._tags:
            r["tags"] = self._tags

        return r


class AtlasChangeSource(AtlasSource):
    """
    Class responsible for creating an Atlas source object for changing
    participants probes for a measurement.
    Usage:
        from ripe.atlas.cousteau import AtlasChangeSource
        as = AtlasChangeSource(**{"type":"area", "value": "WW", "requested": 5})
    """
    def __init__(self, **kwargs):
        if "action" in kwargs:
            self.action = kwargs["action"]
        else:
            self._action = None

        super(AtlasChangeSource, self).__init__(**kwargs)

    # type attribute
    def get_type(self):
        """Getter for type attribute"""
        return self._type

    def set_type(self, value):
        """Setter for type attribute"""
        if self.action == "remove" and value != "probes":
            log = "Sources field 'type' when action is remove should always be 'probes'."
            raise MalFormattedSource(log)
        self._type = value

    doc_type = "Defines the type of probe's source."
    type = property(get_type, set_type, doc=doc_type)

    # tags attribute
    def get_tags(self):
        """Getter for tags attribute"""
        return self._tags

    def set_tags(self, value):
        """Setter for tags attribute"""
        if self.action == "remove":
            log = (
                "Tag-based filtering can only be used when adding "
                "participant probes for a measurement."
            )
            raise MalFormattedSource(log)
        super(AtlasChangeSource, self).set_tags(value)

    doc_tags = "Defines optional tags to filter probes."
    tags = property(get_tags, set_tags, doc=doc_tags)

    # action attribute
    def get_action(self):
        """Getter for action attribute"""
        return self._action

    def set_action(self, value):
        """Setter for action attribute"""
        if value not in ("remove", "add"):
            log = "Sources field 'action' should be 'remove' or 'add'."
            raise MalFormattedSource(log)
        self._action = value

    doc_action = "Defines the action (remove/add if the change source)."
    action = property(get_action, set_action, doc=doc_action)

    def clean(self):
        """
        Cleans/checks user has entered all required attributes. This might save
        some queries from being sent to server if they are totally wrong.
        """
        if not all([self._type, self._requested, self._value, self._action]):
            raise MalFormattedSource(
                "<type, requested, value, action> fields are required."
            )

    def build_api_struct(self):
        """
        Calls parent's method and just adds the addtional field 'action', that
        is required to form the structure that Atlas API is accepting.
        """
        data = super(AtlasChangeSource, self).build_api_struct()
        data.update({"action": self._action})
        return data


class MalFormattedSource(Exception):
    """Custom Exception class for malformed sources"""
    pass

__all__ = ["AtlasSource", "AtlasChangeSource"]



#!/usr/bin/env python
from os.path import abspath, dirname, join
from setuptools import setup

# import manually __version__ variable
__version__ = None
exec(open('ripe/atlas/cousteau/version.py').read())

install_requires = [
    "python-dateutil", "socketIO-client>=0.6.5", "requests>=2.7.0"]

tests_require = [
    "pbr<=1.6",
    "nose",
    "coverage",
    "mock",
    "jsonschema"
]


# Get proper long description for package
current_dir = dirname(abspath(__file__))
description = open(join(current_dir, "README.rst")).read()
changes = open(join(current_dir, "CHANGES.rst")).read()
long_description = '\n\n'.join([description, changes])

setup(
    name="ripe.atlas.cousteau",
    version=__version__,
    packages=["ripe", "ripe.atlas", "ripe.atlas.cousteau"],
    namespace_packages=["ripe", "ripe.atlas"],
    include_package_data=True,
    license="GPLv3",
    url="https://github.com/RIPE-NCC/ripe-atlas-cousteau",
    description="Python wrapper for RIPE Atlas API",
    long_description=long_description,
    author="RIPE Atlas Development Team",
    author_email="atlas-dev@ripe.net",
    maintainer="Andreas Strikos",
    maintainer_email="astrikos@ripe.net",
    install_requires=install_requires,
    tests_require=tests_require,
    test_suite="nose.collector",
    keywords=['RIPE', 'RIPE NCC', 'RIPE Atlas'],
    classifiers=[
        "Operating System :: POSIX",
        "Operating System :: Unix",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2.6",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: Implementation :: CPython",
        "Programming Language :: Python :: Implementation :: PyPy",
        "Topic :: Internet :: WWW/HTTP",
        "Topic :: Software Development :: Libraries :: Python Modules"
    ]
)



"""
Module containing classes responsible for creating different kind of Atlas
measurement objects.
"""


class AtlasMeasurement(object):
    """
    Parent class for creating an Atlas measurement object containing all
    needed options for ATLAS API. The different kind of measurements are
    specified as child classes. These objects can be passed as measurement
    arguments later on when we call AtlasRequest.
    Usage:
        from ripe.atlas.cousteau import Ping
        ping = Ping(**{
            "target": "www.google.gr", "af": 4,
            "description": "testing new wrapper"
        })
    """

    def __init__(self, **kwargs):
        # set to store all options that are used
        self.used_options = set()
        # required options for definitions part
        self.required_options = ["description", "af"]
        self.measurement_type = ""

    def _init(self, **kwargs):
        """
        Initializing required options and set them as attributes as well as
        options coming from user.
        """
        self._init_required_options(**kwargs)
        self.add_option(**kwargs)

    def _store_option(self, option):
        """
        Store option in the used option set. This way we can keep track which
        options user has select to add to instance. This set is used at the
        build_api_struct function when we build the desired data structure from
        user's input.
        """
        self.used_options.add(option)

    def add_option(self, **options):
        """
        Adds an option and its value to the class as an attribute and stores it
        to the used options set.
        """
        for option, value in options.items():
            setattr(self, option, value)
            self._store_option(option)

    def _init_required_options(self, **kwargs):
        """
        Initialize the required option as class members. The value will be
        either None or the specified value in the kwargs or __init__. The logic
        here is to make the required options accesible to edit after a class
        instance has been created.
        """
        for field in self.required_options:
            setattr(self, field, kwargs.get(field))
            self._store_option(field)

    def clean(self):
        """
        Cleans/checks user entered data making sure required options are at
        least present. This might save some queries from being sent if
        they are totally wrong.
        """

        # make sure the correct measurement type is set.
        if not self.measurement_type:
            log = "Please define a valid measurement type."
            raise MalFormattedMeasurement(log)

        # make sure the required fields are set.
        for roption in self.required_options:
            if getattr(self, roption, None) is None:
                log = "%s Measurement field: <%s> is required" % (
                    self.__class__.__name__, roption
                )
                raise MalFormattedMeasurement(log)

    def v2_translator(self, option):
        """
        This is a temporary function that helps move from v1 API to v2 without
        breaking already running script and keep backwards compatibility.
        Translates option name from API v1 to renamed one of v2 API.
        """
        new_option = option
        new_value = getattr(self, option)

        renaming_pairs = {
            "dontfrag": "dont_fragment",
            "maxhops": "max_hops",
            "firsthop": "first_hop",
            "use_NSID": "set_nsid_bit",
            "cd": "set_cd_bit",
            "do": "set_do_bit",
            "qbuf": "include_qbuf",
            "recursion_desired": "set_rd_bit",
            "noabuf": "include_abuf"
        }
        if option in renaming_pairs.keys():
            warninglog = (
                "DeprecationWarning: {0} option has been deprecated and "
                "renamed to {1}."
            ).format(option, renaming_pairs[option])
            print(warninglog)
            new_option = renaming_pairs[option]

        # noabuf was changed to include_abuf so we need a double-negative
        if option == "noabuf":
            new_value = not new_value

        return new_option, new_value

    def build_api_struct(self):
        """
        Calls the clean method of the class and returns the info in a
        structure that Atlas API is accepting.
        """
        self.clean()
        data = {"type": self.measurement_type}

        # add all options
        for option in self.used_options:
            option_key, option_value = self.v2_translator(option)
            data.update({option_key: option_value})

        return data


class Ping(AtlasMeasurement):
    """Class for creating a ping measurement"""

    def __init__(self, **kwargs):
        super(Ping, self).__init__(**kwargs)
        self.measurement_type = "ping"
        self.required_options.extend(["target"])
        self._init(**kwargs)


class Traceroute(AtlasMeasurement):
    """Class for creating a traceroute measurement"""

    def __init__(self, **kwargs):
        super(Traceroute, self).__init__(**kwargs)
        self.measurement_type = "traceroute"
        self.required_options.extend(["target", "protocol"])
        self._init(**kwargs)


class Dns(AtlasMeasurement):
    """Class for creating a DNS measurement"""

    def __init__(self, **kwargs):
        super(Dns, self).__init__(**kwargs)
        self.measurement_type = "dns"
        self.required_options.extend(
            ["query_class", "query_type", "query_argument"]
        )
        self._init(**kwargs)


class Sslcert(AtlasMeasurement):
    """Class for creating an SSL certificate measurement"""

    def __init__(self, **kwargs):
        super(Sslcert, self).__init__(**kwargs)
        self.measurement_type = "sslcert"
        self.required_options.extend(["target"])
        self._init(**kwargs)


class Ntp(AtlasMeasurement):
    """Class for creating an NTP measurement"""

    def __init__(self, **kwargs):
        super(Ntp, self).__init__(**kwargs)
        self.measurement_type = "ntp"
        self.required_options.extend(["target"])
        self._init(**kwargs)


class Http(AtlasMeasurement):
    """Class for creating an HTTP measurement"""

    def __init__(self, **kwargs):
        super(Http, self).__init__(**kwargs)
        self.measurement_type = "http"
        self.required_options.extend(["target"])
        self._init(**kwargs)


class MalFormattedMeasurement(Exception):
    pass

__all__ = ["Ping", "Traceroute", "Dns", "Sslcert", "Ntp", "Http"]



try:
    # Python 3
    from urllib.parse import urlparse
except ImportError:
    # Python 2
    from urlparse import urlparse

import calendar

from datetime import datetime

from .measurement import Ping, Traceroute, Dns, Sslcert, Ntp, Http
from .source import AtlasSource, AtlasChangeSource
from .request import (
    AtlasRequest,
    AtlasCreateRequest,
    AtlasChangeRequest,
    AtlasStopRequest,
    AtlasLatestRequest,
    AtlasResultsRequest
)
from .stream import AtlasStream
from .exceptions import APIResponseError, CousteauGenericError


class EntityRepresentation(object):
    """
    A crude representation of entity's meta data as we get it from the API.
    """

    API_META_URL = ""

    def __init__(self, **kwargs):

        self.id = kwargs.get("id", None)
        self.api_key = kwargs.get("key", "")
        self.meta_data = kwargs.get("meta_data", None)

        if self.meta_data is None and self.id is None:
            raise CousteauGenericError(
                "Id or meta_data should be passed in order to create object."
            )

        if self.meta_data is None:
            if not self._fetch_meta_data():
                raise APIResponseError(self.meta_data)

        self._populate_data()

    def _fetch_meta_data(self):
        """Makes an API call to fetch meta data for the given probe and stores the raw data."""
        is_success, meta_data = AtlasRequest(
            **{"url_path": self.API_META_URL.format(self.id), "key": self.api_key}
        ).get()

        self.meta_data = meta_data
        if not is_success:
            return False

        return True

    def _populate_data(self):
        """Assing some raw meta data from API response to instance properties"""
        raise NotImplementedError()

    def __str__(self):
        return "Probe #{0}".format(self.id)

    def __repr__(self):
        return str(self)


class Probe(EntityRepresentation):
    """
    A crude representation of probe's meta data as we get it from the API.
    """
    API_META_URL = "/api/v2/probes/{0}/"

    def _populate_data(self):
        """Assing some probe's raw meta data from API response to instance properties"""
        if self.id is None:
            self.id = self.meta_data.get("id")
        self.is_anchor = self.meta_data.get("is_anchor")
        self.country_code = self.meta_data.get("country_code")
        self.description = self.meta_data.get("description")
        self.is_public = self.meta_data.get("is_public")
        self.asn_v4 = self.meta_data.get("asn_v4")
        self.asn_v6 = self.meta_data.get("asn_v6")
        self.address_v4 = self.meta_data.get("address_v4")
        self.address_v6 = self.meta_data.get("address_v6")
        self.prefix_v4 = self.meta_data.get("prefix_v4")
        self.prefix_v6 = self.meta_data.get("prefix_v6")
        self.geometry = self.meta_data.get("geometry")
        self.tags = self.meta_data.get("tags")
        self.status = self.meta_data.get("status", {}).get("name")


class Measurement(EntityRepresentation):
    """
    A crude representation of measurement's meta data as we get it from the API.
    """
    API_META_URL = "/api/v2/measurements/{0}/"

    def _populate_data(self):
        """Assing some measurement's raw meta data from API response to instance properties"""
        if self.id is None:
            self.id = self.meta_data.get("id")

        stop_time = self.meta_data.get("stop_time")
        if stop_time:
            stop_time = datetime.fromtimestamp(stop_time)

        self.protocol = self.meta_data.get("af")
        self.destination_address = self.meta_data.get("destination_address")
        self.destination_asn = self.meta_data.get("destination_asn")
        self.destination_name = self.meta_data.get("destination_name")
        self.description = self.meta_data.get("description")
        self.is_oneoff = self.meta_data.get("is_oneoff")
        self.is_public = self.meta_data.get("is_public")
        self.interval = self.meta_data.get("interval")
        self.resolve_on_probe = self.meta_data.get("resolve_on_probe")
        self.creation_time = datetime.fromtimestamp(self.meta_data.get("creation_time"))
        self.start_time = datetime.fromtimestamp(self.meta_data.get("start_time"))
        self.stop_time = stop_time
        self.status_id = self.meta_data.get("status", {}).get("id")
        self.status = self.meta_data.get("status", {}).get("name")
        self.type = self.meta_data.get("type", {}).get("name").upper()
        self.result_url = self.meta_data.get("result")


class RequestGenerator(object):
    """
    Python generator class that yields results for meta APIs like
    probes/measurements as single objects. It supports any filter APIs support
    in a dummy way, which means it will take accept whatever it passed and
    build url_path from this.
    """

    url = ""
    id_filter = ""
    URL_LENGTH_LIMIT = 5000

    def __init__(self, return_objects=False, **filters):
        self.api_filters = filters
        self.split_urls = []
        self.total_count_flag = False
        self.current_batch = []
        self._count = []
        self.return_objects = return_objects
        self.atlas_url = self.build_url()

    def build_url(self):
        """Build the url path based on the filter options."""

        if not self.api_filters:
            return self.url

        # Reduce complex objects to simpler strings
        for k, v in self.api_filters.items():
            if isinstance(v, datetime):  # datetime > UNIX timestamp
                self.api_filters[k] = int(calendar.timegm(v.timetuple()))
            if isinstance(v, (tuple, list)):  # tuples & lists > x,y,z
                self.api_filters[k] = ",".join([str(_) for _ in v])

        if (
            self.id_filter in self.api_filters and
            len(str(self.api_filters[self.id_filter])) > self.URL_LENGTH_LIMIT
        ):
            self.build_url_chunks()
            return self.split_urls.pop(0)

        filters = '&'.join("%s=%s" % (k, v) for (k, v) in self.api_filters.items())

        return "%s?%s" % (self.url, filters)

    def build_url_chunks(self):
        """
        If url is too big because of id filter is huge, break id and construct
        several urls to call them in order to abstract this complexity from user.
        """
        CHUNK_SIZE = 500

        id_filter = str(self.api_filters.pop(self.id_filter)).split(',')
        chuncks = list(self.chunks(id_filter, CHUNK_SIZE))
        filters = '&'.join("%s=%s" % (k, v) for (k, v) in self.api_filters.items())

        for chunk in chuncks:
            if filters:
                url = "{0}?{1}&{2}={3}".format(self.url, filters, self.id_filter, ','.join(chunk))
            else:
                url = "{0}?{1}={2}".format(self.url, self.id_filter, ','.join(chunk))
            self.split_urls.append(url)

    def chunks(self, l, n):
        """Yield successive n-sized chunks from l."""
        for i in range(0, len(l), n):
            yield l[i:i + n]

    def __iter__(self):
        return self

    # Python 3 compatibility
    def __next__(self):
        return self.next()

    def next(self):
        if not self.current_batch:  # If first time or current batch was all given
            if not self.atlas_url:  # We don't have any next url any more, exit
                raise StopIteration()
            self.next_batch()
            if not self.current_batch:  # Server request gives empty batch, exit
                raise StopIteration()

        current_object = self.current_batch.pop(0)
        if self.return_objects:
            return self.object_class(meta_data=current_object)
        else:
            return current_object

    def next_batch(self):
        """
        Querying API for the next batch of objects and store next url and
        batch of objects.
        """
        is_success, results = AtlasRequest(**{"url_path": self.atlas_url}).get()

        if not is_success:
            raise APIResponseError(results)

        self.total_count = results.get("count")
        self.atlas_url = self.build_next_url(results.get("next"))
        self.current_batch = results.get("results", [])

    def build_next_url(self, url):
        """Builds next url in a format compatible with cousteau. Path + query"""
        if not url:
            if self.split_urls:  # If we had a long request give the next part
                self.total_count_flag = False  # Reset flag for count
                return self.split_urls.pop(0)
            else:
                return None

        parsed_url = urlparse(url)
        return "{0}?{1}".format(parsed_url.path, parsed_url.query)

    # count attribute to deal with split-up urls and total count
    def get_total_count(self):
        """Getter for count attribute"""
        if not self._count:
            return 0
        else:
            return sum(self._count)

    def set_total_count(self, value):
        """Setter for count attribute. Set should append only one count per splitted url."""
        if not self.total_count_flag and value:
            self._count.append(int(value))
            self.total_count_flag = True

    doc_count = "Defines how many objects returned."
    total_count = property(get_total_count, set_total_count, doc=doc_count)


class ProbeRequest(RequestGenerator):
    """
    Python generator for Probes meta api.
    e.g.
    for probe in ProbeRequest(**{"limit":200, "country_code": "GR"}):
        print(probe["id"])
    """
    url = "/api/v2/probes/"
    id_filter = "id__in"
    object_class = Probe


class MeasurementRequest(RequestGenerator):
    """
    Python generator for Measurement meta api.
    e.g.
    for measurement in MeasurementRequest(**{"status": 1}):
        print(measurement["id"])
    """
    url = "/api/v2/measurements/"
    id_filter = "id__in"
    object_class = Measurement


__all__ = [
    "Ping",
    "Traceroute",
    "Dns",
    "Sslcert",
    "Ntp",
    "Http",
    "AtlasRequest",
    "AtlasChangeRequest",
    "AtlasCreateRequest",
    "AtlasStopRequest",
    "AtlasLatestRequest",
    "AtlasResultsRequest",
    "AtlasSource",
    "AtlasChangeSource",
    "AtlasStream",
]
